import os
from pathlib import Path

from setuptools import Extension
from loguru import logger

from .activity import get_activity, BuildActivity, DevelopActivity
from .copyutils import copy_directory_contents

class CMakeExtension(Extension):
    def __init__(self, name, source_dir: Path, install_prefix: Path, editable: bool, **kwargs):
        Extension.__init__(self, name, sources=[], **kwargs)
        self.source_dir: Path = source_dir
        self.install_prefix: Path = install_prefix
        self.editable = editable

    def build(self, ext_path: Path, ext_dir: Path) -> None:
        logger.debug(f'ext_path: {ext_path}')
        logger.debug(f'ext_dir: {ext_dir}')

        activity = get_activity()
        logger.debug(activity.__dict__)

        artifacts_dir = activity.artifacts_dir
        logger.debug(f'artifacts_dir: {artifacts_dir}')
        install_from = artifacts_dir / self.install_prefix
        logger.debug(f'install_from: {install_from}')
        install_to = ext_dir
        logger.debug(f'install_to: {install_to}')
        copy_directory_contents(install_from, install_to)
        if self.editable:
            install_editable_to = self.source_dir / self.install_prefix
            logger.debug(f'install_editable_to: {install_editable_to}')
            copy_directory_contents(install_from, install_editable_to)



from typing import Any
import os
from enum import Enum
from pathlib import Path
import json

import pydantic


class BuildMode(str, Enum):
    DEBUG = 'DEBUG'
    RELEASE = 'RELEASE'
    PROFILE = 'PROFILE'


class ActivityType(str, Enum):
    ConfigureActivity = 'ConfigureActivity'
    BuildActivity = 'BuildActivity'
    DevelopActivity = 'DevelopActivity'
    InstallActivity = 'InstallActivity'


class Activity(pydantic.BaseModel):
    type: ActivityType = None
    root: Path = Path.cwd()
    path: Path = Path.cwd() / '_cxbuild/activity.json'
    mode: BuildMode = BuildMode.RELEASE

    def __new__(cls, *args, **kwargs):
        global _activity
        _activity = super(Activity, cls).__new__(cls)
        return _activity

    @property
    def artifacts_dir(self):
        return self.root / '_cxbuild/artifacts'

    # Note:  I'm tempted to serialize to the environment itself instead of a file ...
    def save(self):
        os.environ['CBX_ACTIVITY'] = str(self.path)
        """Serialize an activity to a JSON string"""
        with open(self.path, 'w') as f:
            json.dump({'type': self.type.value, 'object': self.json()}, f)
        return self

_activity: Activity = None


class ConfigureActivity(Activity):
    type: ActivityType = ActivityType.ConfigureActivity


class BuildActivity(Activity):
    type: ActivityType = ActivityType.BuildActivity

class DevelopActivity(Activity):
    type: ActivityType = ActivityType.DevelopActivity
    mode: BuildMode = BuildMode.DEBUG


class InstallActivity(Activity):
    type: ActivityType = ActivityType.InstallActivity

def deserialize_activity(path: Path):
    """Deserialize an activity from a JSON file"""
    with open(path, "r") as f:
        data = json.load(f)

    if data["type"] == ActivityType.ConfigureActivity.value:
        return ConfigureActivity.parse_raw(data["object"])
    elif data["type"] == ActivityType.BuildActivity.value:
        return BuildActivity.parse_raw(data["object"])
    elif data["type"] == ActivityType.DevelopActivity.value:
        return DevelopActivity.parse_raw(data["object"])
    else:
        raise ValueError("Invalid type")


def get_activity() -> Activity:
    global _activity
    if _activity:
        return _activity
    path = Path(os.environ["CBX_ACTIVITY"])
    return deserialize_activity(path)



from pathlib import Path

from .solution import Solution

import socket
from loguru import logger

HOST, PORT = 'localhost', 5005

sock = socket.socket(socket.AF_INET, # Internet
                     socket.SOCK_DGRAM) # UDP

def logudp(message: str):
    sock.sendto(bytes(message, 'utf-8'), (HOST, PORT))

logger.add(logudp)
#logger.add("cxbuild_{time}.log")

class CxBuild:
    def __init__(self) -> None:
        self.solution = Solution(Path.cwd())

    def clean(self):
        print('clean')
        self.solution.clean()

    def configure(self):
        print('configure')
        self.solution.configure()

    def develop(self, project_name: str = None):
        print('develop')
        #self.configure()
        self.solution.develop(project_name)

    def build(self):
        print('build')
        #self.configure()
        self.solution.build()

    def install(self):
        print('install')
        self.solution.install()



from pathlib import Path
import tomli

class Config:
    def __init__(self, config) -> None:
        self.config = config
        for key, value in config.items():
            if isinstance(value, dict):
                value = Config(value)
            setattr(self, key, value)

class PyProject(Config):
    def __init__(self, config) -> None:
        super().__init__(config)

    @classmethod
    def load(cls, path: Path):
        path = path / 'pyproject.toml'
        with open(path, "rb") as f:
            config = tomli.load(f)
            return PyProject(config)

    """
    @property
    def project(self):
        return self.config["project"]

    @property
    def tool(self):
        return self.config["tool"]
    """
    @property
    def name(self):
        #return self.project["name"]
        return self.project.name



from pathlib import Path

from setuptools.command.build_ext import build_ext

from .cmake_extension import CMakeExtension

class ExtensionBuilder(build_ext):

    def initialize_options(self):
        build_ext.initialize_options(self)


    def finalize_options(self):
        build_ext.finalize_options(self)

    def run(self) -> None:
        cmake_extensions = [e for e in self.extensions if isinstance(e, CMakeExtension)]

        if len(cmake_extensions) == 0:
            raise ValueError("No CMakeExtension objects found")

        for ext in cmake_extensions:
            self.build_extension(ext)

    def build_extension(self, ext: CMakeExtension) -> None:
        ext_path = self.get_ext_fullpath(ext.name)
        ext_dir = Path(ext_path).parent.absolute()
        ext.build(ext_path, ext_dir)


import sys, os
from pathlib import Path
from loguru import logger
import setuptools
from setuptools import Distribution
import shutil

from .cmake_extension import CMakeExtension
from .extension_builder import ExtensionBuilder

from .activity import get_activity, BuildActivity, DevelopActivity
from .project_base import ProjectBase
from .build_tool import BuildConfig, BuildTool
from .pip_tool import PipConfig, PipTool
from .copyutils import copy_directory_contents

class Project(ProjectBase):
    def __init__(self, path: Path) -> None:
        super().__init__(path)

    def clean(self):
        logger.debug('clean')
        dist_dir = self.path / 'dist'
        if dist_dir.exists():
            logger.debug(f"removing {dist_dir}")
            shutil.rmtree(dist_dir)

        build_dir = self.path / 'build'
        if build_dir.exists():
            logger.debug(f"removing {build_dir}")
            shutil.rmtree(build_dir)

        for egginfo_dir in self.path.rglob('*.egg-info'):
            if egginfo_dir.is_dir():
                logger.debug(f"removing {egginfo_dir}")
                shutil.rmtree(egginfo_dir)

    def develop(self):
        logger.debug('develop')
        tool = PipTool(PipConfig(env=os.environ, source_dir=self.path))
        tool.install()

    def build(self):
        tool = BuildTool(BuildConfig(env=os.environ, source_dir=self.path))
        tool.build()

    def build_wheel(
        self,
        wheel_directory: str,
        config_settings: dict[str, list[str] | str] | None = None,
        metadata_directory: str | None = None
    ) -> str:
        logger.debug('build_wheel')
        name: str = self.pyproject.tool.cxbuild.extension.name
        logger.debug(f'extension name: {name}')
        split_name = name.split('.')
        split_name.pop()
        install_prefix = Path(*split_name)
        logger.debug(f'install prefix{install_prefix}')
        
        activity = get_activity()
        logger.debug(f'activity: {activity.__dict__}')

        editable = True if isinstance(activity, DevelopActivity) else False

        dist_dir = self.path / 'dist'
        '''
        if dist_dir.exists():
            shutil.rmtree(dist_dir)
        '''

        #Note:  This is a hack, but I didn't want to call yet another subprocess
        sys.argv = ["setup.py", "bdist_wheel"]
        dist: Distribution = setuptools.setup(
            ext_modules=[
                CMakeExtension(
                    name=name,
                    source_dir=Path.cwd(),
                    install_prefix=install_prefix,
                    editable=editable
                ),
            ],
            cmdclass=dict(
                build_ext=ExtensionBuilder,
            )
        )

        out_dir = Path(wheel_directory)
        items = list(dist_dir.iterdir())
        dist_name = items[0].name
        logger.debug(dist_name)

        if dist_dir != out_dir:
            copy_directory_contents(dist_dir, out_dir)

        return dist_name

    def write_requirements(self, requirements):
        with open(self.path / 'requirements.txt', 'w') as f:
            f.write('\n'.join(requirements))


from pathlib import Path

from .pyproject import PyProject

class ProjectBase:
    def __init__(self, path: Path) -> None:
        self.path: Path = path
        self.pyproject:PyProject = PyProject.load(path)
        self.name: str = self.pyproject.name


import subprocess


class Tool:
    def __init__(self) -> None:
        pass

    def run(self, cmd: list):
        #subprocess.check_call(cmd)
        subprocess.run(cmd)
        """
        #cwd = self.config.source_dir
        #env = self.config.env
        try:
            #result = subprocess.run(cmd, cwd=cwd, env=env, check=True, text=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
            result = subprocess.run(cmd)
            print("Command executed successfully. Output:\n", result.stdout)
        except subprocess.CalledProcessError as e:
            print(f"Command execution failed with error code {e.returncode}. Error message:\n", e.stderr)
        """


from __future__ import annotations

import dataclasses
import sys, os
import shutil
from pathlib import Path

from .tool import Tool


class CMakeConfigError(Exception):
    """
    Something is misconfigured.
    """

@dataclasses.dataclass
class CMakeConfig:
    source_dir: Path
    build_dir: Path
    build_type: str
    generator: str

    module_dirs: list[Path] = dataclasses.field(default_factory=list)
    prefix_dirs: list[Path] = dataclasses.field(default_factory=list)
    init_cache_file: Path = dataclasses.field(init=False, default=Path())
    env: dict[str, str] = dataclasses.field(init=False, default_factory=os.environ.copy)
    single_config: bool = not sys.platform.startswith("win32")

    def __post_init__(self) -> None:
        self.init_cache_file = self.build_dir / "CMakeInit.txt"

        if not self.source_dir.is_dir():
            msg = f"source directory {self.source_dir} does not exist"
            raise CMakeConfigError(msg)

        self.build_dir.mkdir(parents=True, exist_ok=True)
        if not self.build_dir.is_dir():
            msg = f"build directory {self.build_dir} must be a (creatable) directory"
            raise CMakeConfigError(msg)

def join_posix_paths(paths: list[Path]):
    posix_paths = ";".join(str(path.as_posix()) for path in paths)
    print(posix_paths)
    return posix_paths

class CMakeTool(Tool):
    def __init__(self, config: CMakeConfig) -> None:
        super().__init__()
        self.config = config

    def configure(self):
        print('configure')
        # Initialize the CMake configuration arguments
        configure_args = []

        # Select the appropriate generator and accompanying settings
        if self.config.generator is not None:
            configure_args += [f'-G "{self.config.generator}"']

            if self.config.generator == "Ninja":
                configure_args += [f"-DCMAKE_MAKE_PROGRAM={shutil.which('ninja')}"]

        # CMake configure arguments
        cmake_install_prefix = Path.cwd() / '_cxbuild/artifacts'

        cmake_prefix_path = join_posix_paths(self.config.prefix_dirs)
        print('cmake_prefix_path: ', cmake_prefix_path)

        configure_args += [
            f'-DCMAKE_BUILD_TYPE={self.config.build_type}',
            f'-DCMAKE_INSTALL_PREFIX:PATH={cmake_install_prefix}',
            #f"-DCMAKE_MODULE_PATH:PATH={cmake_module_path}",
            f'-DCMAKE_PREFIX_PATH:PATH={cmake_prefix_path}',
        ]

        command = [
            "cmake",
            "-S",
            self.config.source_dir,
            "-B",
            self.config.build_dir,
        ] + configure_args

        self.run(command)

    def build(self):
        print('build')
        build_args = ["--config", self.config.build_type]
        command = ["cmake", "--build", self.config.build_dir] + build_args
        self.run(command)

    def install(self):
        print('install')
        command = ["cmake", "--install", self.config.build_dir]
        self.run(command)


import os
import fnmatch

import shutil
from pathlib import Path

def copy_directory_contents(src: Path, dest: Path):
    # Create the destination directory if it doesn't exist
    dest.mkdir(parents=True, exist_ok=True)

    # Iterate through the source directory contents and copy each item to the destination directory
    for item in src.iterdir():
        if item.is_file():
            shutil.copy2(item, dest / item.name)
        elif item.is_dir():
            # Recursively copy subdirectories
            copy_directory_contents(item, dest / item.name)

def recursive_copy(src: Path, dst: Path, exclude_files=[], exclude_dirs=[]):
    src = Path(src)
    dst = Path(dst)
    
    if not src.is_dir():
        raise ValueError(f"Source {src} is not a directory")
    if not dst.is_dir():
        raise ValueError(f"Destination {dst} is not a directory")
    if src == dst:
        raise Exception("Source and Destination directories are the same!")

    for item in src.iterdir():
        # If the item is a directory and does not match any of the exclude patterns
        if item.is_dir() and not any(fnmatch.fnmatch(item, pattern) for pattern in exclude_dirs):
            new_dst = dst / item.name
            new_dst.mkdir(exist_ok=True)
            recursive_copy(item, new_dst, exclude_files, exclude_dirs)
        # If the item is a file and does not match any of the exclude patterns
        elif item.is_file() and not any(fnmatch.fnmatch(item, pattern) for pattern in exclude_files):
            shutil.copy(item, dst)

"""
def recursive_copy(src, dst, exclude_files=None, exclude_dirs=None):
    if exclude_files is None:
        exclude_files = []
    if exclude_dirs is None:
        exclude_dirs = []

    if not os.path.exists(dst):
        os.makedirs(dst)

    for item in os.listdir(src):
        src_item = os.path.join(src, item)
        dst_item = os.path.join(dst, item)

        if os.path.isdir(src_item):
            # Check if the directory matches any of the exclude patterns
            if not any(fnmatch.fnmatch(item, pattern) for pattern in exclude_dirs):
                recursive_copy(src_item, dst_item, exclude_files, exclude_dirs)
        else:
            # Check if the file matches any of the exclude patterns
            if not any(fnmatch.fnmatch(item, pattern) for pattern in exclude_files):
                shutil.copy2(src_item, dst_item)
"""

# Example usage:
"""
src = 'source_directory'
dst = 'destination_directory'
exclude_files = ['*.tmp', '*.log']
exclude_dirs = ['__pycache__', 'temp']

recursive_copy(src, dst, exclude_files, exclude_dirs)
"""


from __future__ import annotations

import dataclasses
from pathlib import Path

from .tool import Tool
import subprocess


class PipConfigError(Exception):
    """
    Something is misconfigured.
    """


@dataclasses.dataclass
class PipConfig:
    env: dict[str, str]
    source_dir: Path


class PipTool(Tool):
    def __init__(self, config: PipConfig = None) -> None:
        super().__init__()
        self.config = config

    def install(self):
        cmd = ["python", "-m", "pip", "install", "--no-build-isolation", "--editable", "."]
        cwd = self.config.source_dir
        env = self.config.env
        try:
            result = subprocess.run(cmd, cwd=cwd, env=env, check=True, text=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
            print("Command executed successfully. Output:\n", result.stdout)
        except subprocess.CalledProcessError as e:
            print(f"Command execution failed with error code {e.returncode}. Error message:\n", e.stderr)



"""
PEP 517 build hooks
"""


from __future__ import annotations
from typing import Mapping, Any
import os
import setuptools.build_meta as build_meta

__all__ = [
    "_supported_features",
    "build_sdist",
    "build_wheel",
    "build_editable",
    "get_requires_for_build_sdist",
    "get_requires_for_build_wheel",
    "get_requires_for_build_editable",
    "prepare_metadata_for_build_wheel",
    "prepare_metadata_for_build_editable",
]

from pathlib import Path
from loguru import logger

from .project import Project


def _supported_features():
    return ["build_editable"]


def build_sdist(
    sdist_directory: str,
    config_settings: dict[str, list[str] | str] | None = None,
) -> str:
    logger.debug("Build hook: build_sdist")
    return build_meta.build_sdist(sdist_directory, config_settings)


def build_wheel(
    wheel_directory: str,
    config_settings: dict[str, list[str] | str] | None = None,
    metadata_directory: str | None = None,
) -> str:
    logger.debug("Build hook: build_wheel")
    logger.debug(f"wheel_directory: {wheel_directory}")
    project = Project(Path.cwd())
    return project.build_wheel(wheel_directory, config_settings, metadata_directory)


def build_editable(
    wheel_directory: str,
    config_settings: dict[str, list[str] | str] | None = None,
    metadata_directory: str | None = None,
) -> str:
    logger.debug("Build hook: build_editable")
    logger.debug(f"wheel_directory: {wheel_directory}")
    # If not invoked indirectly by cxbuild itself do the default action
    if not os.environ.get('CBX_ACTIVITY'):
        return build_meta.build_editable(wheel_directory, config_settings, metadata_directory)
    return build_wheel(wheel_directory, config_settings, metadata_directory)


def get_requires_for_build_sdist(
    config_settings: dict[str, str | list[str]] | None = None  # noqa: ARG001
) -> list[str]:
    logger.debug("Build hook: get_requires_for_build_sdist")
    return build_meta.get_requires_for_build_sdist(config_settings)


def get_requires_for_build_wheel(
    config_settings: Mapping[str, Any] | None = None
) -> list[str]:
    logger.debug("Build hook: get_requires_for_build_wheel")
    #return []
    return build_meta.get_requires_for_build_wheel(config_settings)


def get_requires_for_build_editable(self, config_settings=None):
    logger.debug("Build hook: get_requires_for_build_editable")
    return get_requires_for_build_wheel(config_settings)


def prepare_metadata_for_build_wheel(
    metadata_directory: str,
    config_settings: dict[str, list[str] | str] | None = None,
) -> str:
    logger.debug("Build hook: prepare_metadata_for_build_wheel")
    return build_meta.prepare_metadata_for_build_wheel(
        metadata_directory, config_settings
    )


def prepare_metadata_for_build_editable(metadata_directory, config_settings=None):
    logger.debug("Build hook: build_editable")
    return build_meta.prepare_metadata_for_build_wheel(
        metadata_directory, config_settings
    )



import click
from click import Context

from ..cxbuild import CxBuild

@click.group(invoke_without_command=True)
@click.pass_context
def cli(ctx):
    ctx.ensure_object(dict)
    if ctx.invoked_subcommand is None:
        build()


@cli.command()
@click.pass_context
def clean(ctx: Context):
    builder = CxBuild()
    builder.clean()

@cli.command()
@click.pass_context
def configure(ctx: Context):
    builder = CxBuild()
    builder.configure()

@cli.command()
@click.pass_context
@click.argument("project_name", required=False)
def develop(ctx: Context, project_name: str):
    builder = CxBuild()
    builder.develop(project_name)

@cli.command()
@click.pass_context
def build(ctx: Context):
    builder = CxBuild()
    builder.clean()
    builder.build()

@cli.command()
@click.pass_context
def install(ctx: Context):
    builder = CxBuild()
    builder.install()



import site
import importlib
from pathlib import Path

from loguru import logger

from .cmake_tool import CMakeConfig, CMakeTool

from .activity import (
    get_activity,
    BuildMode,
    Activity,
    ConfigureActivity,
    BuildActivity,
    DevelopActivity,
    InstallActivity,
)
from .project_base import ProjectBase
from .project import Project


def is_glob(s):
    return any(char in s for char in "*?[]")


class Solution(ProjectBase):
    def __init__(self, path: Path) -> None:
        super().__init__(path)
        logger.debug(f"solution path: {path}")
        self.build_root = path / "_cxbuild"
        self.ensure()
        self.projects: list[Project] = []
        self.project_map: dict[str, Project] = {}
        self.create_projects()

    def select_projects(self, project_name: str = None):
        if project_name:
            project = self.project_map.get(project_name)
            if project:
                return [project]
            else:
                raise Exception(f"project not found: {project_name}")
        else:
            return self.projects

    def ensure(self):
        if not self.build_root.exists():
            self.build_root.mkdir()

    def create_projects(self):
        if not hasattr(self.pyproject.tool.cxbuild, "projects"):
            return
        project_globs = self.pyproject.tool.cxbuild.projects
        logger.debug(f"project_globs: {project_globs}")
        project_paths = []
        for glob in project_globs:
            if is_glob(glob):
                project_paths += list(self.path.glob(glob))
            else:
                project_paths.append(self.path / glob)
        logger.debug(f"project_paths: {project_paths}")
        for project_path in project_paths:
            self.add_project(Project(project_path))

    def add_project(self, project: Project):
        self.projects.append(project)
        self.project_map[project.name] = project

    def create_config(self, activity: Activity):
        prefix_dirs = []
        site_packages = site.getsitepackages()
        logger.debug(f"site_packages: {site_packages}")
        for site_package in site_packages:
            prefix_dirs.append(Path(site_package))

        if hasattr(self.pyproject.tool.cxbuild, "plugins"):
            plugins = self.pyproject.tool.cxbuild.plugins
            logger.debug(f"plugins: {plugins}")
            for plugin in plugins:
                plugin_module = importlib.import_module(plugin)
                logger.debug(f"plugin_module: {plugin_module}")
                plugin_prefix = Path(plugin_module.__file__).parent.parent
                logger.debug(f"plugin_prefix: {plugin_prefix}")
                prefix_dirs.append(plugin_prefix)

        logger.debug(f"prefix_dirs: {prefix_dirs}")
        build_type = "Release"
        # build_type = 'Debug'

        # TODO:  This is chicken&egg, find a better solution
        if activity.mode == BuildMode.DEBUG:
            build_type = "Debug"

        logger.info(f"Configuring in {build_type} mode")
        config = CMakeConfig(
            source_dir=Path("."),
            build_dir=Path("_cxbuild/build"),
            build_type=build_type,
            generator=None,
            prefix_dirs=prefix_dirs,
        )
        return config

    def clean(self):
        logger.info("clean")
        for project in self.projects:
            project.clean()
        """
        if activity.mode == BuildMode.DEBUG:
            build_dir = self.path / '_cxbuild/build'
            if build_dir.exists():
                logger.debug(f"removing {build_dir}")
                shutil.rmtree(build_dir)
        """

    def configure(self):
        logger.info("configure")
        activity = ConfigureActivity().save()
        config = self.create_config(activity)
        tool = CMakeTool(config)
        tool.configure()

    def develop(self, project_name: str = None):
        logger.info("develop")
        activity = DevelopActivity().save()

        config = self.create_config(activity)
        tool = CMakeTool(config)
        tool.configure()
        tool.build()
        tool.install()

        for project in self.select_projects(project_name):
            project.develop()

    def build(self):
        logger.info("build")
        activity = BuildActivity().save()

        config = self.create_config(activity)
        tool = CMakeTool(config)
        tool.configure()
        tool.build()
        tool.install()

        for project in self.projects:
            project.build()

    def install(self):
        logger.info("install")
        activity = InstallActivity().save()
        config = self.create_config(activity)
        tool = CMakeTool(config)
        tool.install()
